/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
 * FML interpreter. Europagate, 1995
 *
 * $Log: fml.c,v $
 * Revision 1.16  1995/05/16 09:39:32  adam
 * LICENSE.
 *
 * Revision 1.15  1995/02/27  09:01:20  adam
 * Regular expression support. Argument passing by name option. New FML
 * function strlen.
 *
 * Revision 1.14  1995/02/23  08:32:04  adam
 * Changed header.
 *
 * Revision 1.12  1995/02/22  15:20:13  adam
 * Bug fix in fml_exec_space.
 *
 * Revision 1.11  1995/02/22  08:50:49  adam
 * Definition of CPP changed. Output function can be customized.
 *
 * Revision 1.10  1995/02/21  17:46:08  adam
 * Bug fix in fml_sub0.
 *
 * Revision 1.9  1995/02/21  14:00:03  adam
 * Minor changes.
 *
 * Revision 1.8  1995/02/10  18:15:52  adam
 * FML function 'strcmp' implemented. This function can be used to
 * test for existence of MARC fields.
 *
 * Revision 1.7  1995/02/10  15:50:54  adam
 * MARC interface implemented. Minor bugs fixed. fmltest can
 * be used to format single MARC records. New function '\list'
 * implemented.
 *
 * Revision 1.6  1995/02/09  16:06:06  adam
 * FML can be called from the outside multiple times by the functions:
 * fml_exec_call and fml_exec_call_str.
 * An interactive parameter (-i) to fmltest starts a shell-like
 * interface to FML by using the fml_exec_call_str function.
 *
 * Revision 1.5  1995/02/09  14:33:36  adam
 * Split source fml.c and define relevant build-in functions in separate
 * files. New operators mult, div, not, llen implemented.
 *
 * Revision 1.4  1995/02/09  13:07:14  adam
 * Nodes are freed now. Many bugs fixed.
 *
 * Revision 1.3  1995/02/07  16:09:23  adam
 * The \ character is no longer INCLUDED when terminating a token.
 * Major changes in tokenization routines. Bug fixes in expressions
 * with lists (fml_sub0).
 *
 * Revision 1.2  1995/02/06  15:23:25  adam
 * Added some more relational operators (le,ne,ge). Added increment
 * and decrement operators. Function index changed, so that first
 * element is 1 - not 0. Function fml_atom_val edited.
 *
 * Revision 1.1.1.1  1995/02/06  13:48:10  adam
 * First version of the FML interpreter. It's slow and memory isn't
 * freed properly. In particular, the FML nodes aren't released yet.
 *
 */
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "fmlp.h"

static int default_read_func (void)
{
    return getchar ();
}

static void default_write_func (int c)
{
    putchar (c);
}

static void default_err_handle (int no)
{
    fprintf (stderr, "Error: %d\n", no);
}

static struct fml_node *fml_sub0 (Fml fml, struct fml_node *list);
static struct fml_node *fml_sub1 (Fml fml, struct fml_node **lp,
                                  struct token *tp);
static struct fml_node *fml_sub2 (Fml fml, struct fml_node **lp,
                                  struct token *tp);
static struct fml_node *fml_exec_space (Fml fml, struct fml_node **lp, 
                                        struct token *tp);
static struct fml_node *fml_exec_nl (Fml fml, struct fml_node **lp, 
                                     struct token *tp);
                                  
static int indent = 0;

static void pr_indent (int n)
{
    assert (indent >= 0);
    if (n >= 0)
    {
        int i = indent;
        while (--i >= 0)
            putchar(' ');
    }
    if (n > 0)
    {
        printf ("[");
        ++indent;
    } 
    else if (n < 0)
    {
        printf ("]\n");
        --indent;
    }
}

Fml fml_open (void)
{
    struct fml_sym_info *sym_info;

    Fml fml = malloc (sizeof(*fml));

    if (!fml)
        return NULL;

    fml->escape_char = '\\';
    fml->comment_char = '#';
    fml->eof_mark = EOF;
    fml->white_chars = " \t\f\r\n";
    fml->read_func = default_read_func;
    fml->err_handle = default_err_handle;
    fml->write_func = default_write_func;

    fml->list = NULL;
    fml->sym_tab = fml_sym_open ();
    fml->atom_free_list = NULL;
    fml->node_free_list = NULL;
    fml->debug = 0;

    sym_info = fml_sym_add (fml->sym_tab, "func");
    sym_info->kind = FML_FUNC;
    sym_info = fml_sym_add (fml->sym_tab, "bin");
    sym_info->kind = FML_BIN;
    sym_info = fml_sym_add (fml->sym_tab, "if");
    sym_info->kind = FML_IF;
    sym_info = fml_sym_add (fml->sym_tab, "else");
    sym_info->kind = FML_ELSE;
    sym_info = fml_sym_add (fml->sym_tab, "foreach");
    sym_info->kind = FML_FOREACH;
    sym_info = fml_sym_add (fml->sym_tab, "set");
    sym_info->kind = FML_SET;
    sym_info = fml_sym_add (fml->sym_tab, "while");
    sym_info->kind = FML_WHILE;
    sym_info = fml_sym_add (fml->sym_tab, "return");
    sym_info->kind = FML_RETURN;

    fml_list_init (fml);
    fml_arit_init (fml);
    fml_rel_init (fml);
    fml_str_init (fml);

    sym_info = fml_sym_add (fml->sym_tab, "s");
    sym_info->kind = FML_CPREFIX;
    sym_info->prefix = fml_exec_space;
    sym_info = fml_sym_add (fml->sym_tab, " ");
    sym_info->kind = FML_CPREFIX;
    sym_info->prefix = fml_exec_space;
    sym_info = fml_sym_add (fml->sym_tab, "n");
    sym_info->kind = FML_CPREFIX;
    sym_info->prefix = fml_exec_nl;

    return fml;
}

static Fml fml_pop_handler = NULL;
static void pop_handler (struct fml_sym_info *info)
{
    assert (fml_pop_handler);
    switch (info->kind)
    {
    case FML_VAR:
        fml_node_delete (fml_pop_handler, info->body);
        break;
    }
}
static void fml_do_pop (Fml fml)
{
    fml_pop_handler = fml;
    fml_sym_pop (fml->sym_tab, pop_handler);
}

int fml_preprocess (Fml fml)
{
    fml->list = fml_tokenize (fml);
    return 0;
}


void fml_init_token (struct token *tp, Fml fml)
{
    tp->maxbuf = FML_ATOM_BUF*2;
    tp->offset = 0;
    tp->atombuf = tp->sbuf;
    tp->tokenbuf = tp->sbuf + tp->maxbuf;
    tp->escape_char = fml->escape_char;
}

void fml_del_token (struct token *tp, Fml fml)
{
    if (tp->maxbuf != FML_ATOM_BUF*2)
        free (tp->atombuf);
}

void fml_cmd_lex (struct fml_node **np, struct token *tp)
{
    fml_cmd_lex_s (np, tp, 1);
}

void fml_cmd_lex_s (struct fml_node **np, struct token *tp, int esc_stop)
{
    char *cp;
    char *dst;
    if (!*np)
    {
        tp->kind = '\0';
        return;
    }
    if (tp->offset == 0)
    {
        tp->separate = 1;
        if ((*np)->is_atom)
        {
            tp->atom = (*np)->p[0];
            if (!tp->atom->next)
                fml_atom_str (tp->atom, tp->atombuf);
            else
            {
                int l = fml_atom_str (tp->atom, NULL);
                if (l >= tp->maxbuf-1)
                {
                    if (tp->maxbuf != FML_ATOM_BUF*2)
                        free (tp->atombuf);
                    tp->maxbuf = l + 40;
                    tp->atombuf = malloc (tp->maxbuf*2);
                    tp->tokenbuf = tp->atombuf + tp->maxbuf;
                }
                fml_atom_str (tp->atom, tp->atombuf);
            }
        }
        else
        {
            tp->sub = (*np)->p[0];
            tp->kind = 'g';
            *np = (*np)->p[1];
            return ;
        }
    }
    else
        tp->separate = 0;
    cp = tp->atombuf + tp->offset;
    dst = tp->tokenbuf;
    if (*cp == tp->escape_char)
    {
        tp->kind = 'e';
        cp++;
        if (*cp == '\0')
        {
            strcpy (dst, " ");
            tp->offset = 0;
            *np = (*np)->p[1];
            return ;
        }
    }
    else
    {
        tp->kind = 't';
    }
    while (*cp)
    {
        if (*cp == tp->escape_char && esc_stop)
        {
            *dst = '\0';
            tp->offset = cp - tp->atombuf;
            return ;
        }
        *dst++ = *cp++;
    }
    *dst = '\0';
    tp->offset = 0;
    *np = (*np)->p[1];
}

struct fml_node *fml_expr_term (Fml fml, struct fml_node **lp, 
                                struct token *tp)
{
    struct fml_node *fn;
    if (tp->kind == 'g')
    {
        fn = fml_sub0 (fml, tp->sub);
        fml_cmd_lex (lp, tp);
    }
    else
        fn = fml_sub2 (fml, lp, tp);
    return fn;
}

void fml_lr_values (Fml fml, struct fml_node *l, int *left_val,
                           struct fml_node *r, int *right_val)
{
    if (l && l->is_atom)
        *left_val = fml_atom_val (l->p[0]);
    else
        *left_val = 0;
    if (r && r->is_atom)
        *right_val = fml_atom_val (r->p[0]);
    else
        *right_val = 0;
    fml_node_delete (fml, l);
    fml_node_delete (fml, r);
}

static struct fml_node *fml_exec_space (Fml fml, struct fml_node **lp, 
                                        struct token *tp)
{
    fml_cmd_lex (lp, tp);
    if (fml->debug & 1) 
        (*fml->write_func) ('_');
    else
        (*fml->write_func) (' ');
    return NULL;
}

static struct fml_node *fml_exec_nl (Fml fml, struct fml_node **lp, 
                                     struct token *tp)
{
    fml_cmd_lex (lp, tp);
    (*fml->write_func) ('\n');
    return NULL;
}

static struct fml_node *fml_exec_prefix (struct fml_sym_info *info, Fml fml,
                                         struct fml_node **lp,
                                         struct token *tp)
{
    struct fml_node *fn;
    struct fml_sym_info *arg_info;
    struct fml_node *return_value;
    static char arg_name[128];

    if (fml->debug & 1)
    {
        pr_indent (1);
        printf ("exec_prefix ");
    }
    fml_sym_push (fml->sym_tab);
    fml_cmd_lex (lp, tp);
    for (fn = info->args; fn; fn = fn->p[1])
    {
        assert (fn->is_atom);
        fml_atom_strx (fn->p[0], arg_name, 127);
        if (fml->debug & 1)
        {
            pr_indent (1);
            printf ("%s=", arg_name);
        }
        if (*arg_name == fml->escape_char)
        {
            arg_info = fml_sym_add_local (fml->sym_tab, 1+arg_name);
            arg_info->kind = FML_CODE;

            if (tp->kind == 'g')
                arg_info->body = tp->sub;
            else
                arg_info->body = NULL;
            if (fml->debug & 1)
            {
                fml_pr_list (arg_info->body);
                pr_indent (-1);
            }
            fml_cmd_lex (lp, tp);
        }
        else
        {
            arg_info = fml_sym_add_local (fml->sym_tab, arg_name);
            arg_info->kind = FML_VAR;

            if (tp->kind == 'g')
            {
                arg_info->body = fml_sub0 (fml, tp->sub);
                fml_cmd_lex (lp, tp);
            }
            else
                arg_info->body = fml_sub2 (fml, lp, tp);
            if (fml->debug & 1)
            {
                fml_pr_list (arg_info->body);
                pr_indent (-1);
            }
        }
    }
    return_value = fml_exec_group (info->body, fml);
    if (fml->debug & 1)
    {
        pr_indent(0);
        pr_indent (-1);
    }
    fml_do_pop (fml);
    return return_value;
}


static void fml_emit (Fml fml, struct fml_node *list)
{
    int s = 0;
    while (list)
    {
        if (list->is_atom)
        {
            struct fml_atom *a;
            if (s)
                (*fml->write_func) (' ');
            s++;
            for (a = list->p[0]; a; a=a->next)
            {
                int i = 0;
                while (i < FML_ATOM_BUF && a->buf[i])
                    (*fml->write_func) (a->buf[i++]);
            }
        }
        else
            fml_emit (fml, list->p[0]);
        list = list->p[1];
    }
}


static struct fml_node *fml_sub2 (Fml fml, struct fml_node **lp,
                                  struct token *tp)
{
    struct fml_node *fn;
    struct fml_sym_info *info;
    if (tp->kind == 'e')
    {
        info = fml_sym_lookup (fml->sym_tab, tp->tokenbuf);
	if (!info)
        {
	    printf ("<<unknown %s in expression>>", tp->tokenbuf);
	    getchar ();
	    return NULL;
        }
        switch (info->kind)
        {
        case FML_VAR:
            fn = fml_node_copy (fml, info->body);           
            fml_cmd_lex (lp, tp);
            break;
        case FML_CODE:
            fn = fml_node_copy (fml, info->body);           
            fml_cmd_lex (lp, tp);
            break;
        case FML_PREFIX:
            fn = fml_exec_prefix (info, fml, lp, tp);
            break;
        case FML_CPREFIX:
            fn = (*info->prefix) (fml, lp, tp);
            break;
        default:
            fml_cmd_lex (lp, tp);
            fn = NULL;
        }
    }
    else if (tp->kind == 'g')
    {
        if (tp->sub)
            fn = fml_sub0 (fml, tp->sub);
        else
            fn = NULL;
        fml_cmd_lex (lp, tp);
    }
    else if (tp->kind == 't')
    {
        fn = fml_node_alloc (fml);
        fn->is_atom = 1;
        fn->p[0] = fml_atom_alloc (fml, tp->tokenbuf);
        fml_cmd_lex (lp, tp);
    }
    else
        fn = NULL;
    return fn;
}

static struct fml_node *fml_sub1 (Fml fml, struct fml_node **lp,
                                  struct token *tp)
{
    struct fml_node *f1, *f2, *fn;
    struct fml_sym_info *info;

    f1 = fml_sub2 (fml, lp, tp);
    while (tp->kind == 'e')
    {
        info = fml_sym_lookup (fml->sym_tab, tp->tokenbuf);
        if (!info)
        {
            fprintf (stderr, "cannot locate `%s'", tp->tokenbuf);
            exit (1);
        }
        if (info->kind == FML_CBINARY)
        {
            fml_cmd_lex (lp, tp);
            f2 = fml_sub2 (fml, lp, tp);
            fn = (*info->binary) (fml, f1, f2);
            f1 = fn;
            continue;
        }
        else if (info->kind == FML_BINARY)
        {
            struct fml_sym_info *arg_info;
            char arg[127];

            if (fml->debug & 1)
            {
                pr_indent (1);
                printf ("exec binary %s", tp->tokenbuf);
            }
            fml_cmd_lex (lp, tp);
            f2 = fml_sub2 (fml, lp, tp);
            fml_sym_push (fml->sym_tab);

            fml_atom_strx (info->args->p[0], arg, 127);
            arg_info = fml_sym_add_local (fml->sym_tab, arg);
            arg_info->kind = FML_VAR;
            arg_info->body = f1;
            if (fml->debug & 1)
            {
                printf (" left=");
                fml_pr_list (f1);
            }
            fml_atom_strx ( ((struct fml_node *) info->args->p[1])->p[0],
                           arg, 127);
            arg_info = fml_sym_add_local (fml->sym_tab, arg);
            arg_info->kind = FML_VAR;
            arg_info->body = f2;
            if (fml->debug & 1)
            {
                printf (" right=");
                fml_pr_list (f2);
                putchar ('\n');
            }
            f1 = fml_exec_group (info->body, fml);
            fml_do_pop (fml);
            if (fml->debug & 1)
            {
                pr_indent (0);
                pr_indent (-1);
            }
        }
        else
            break;
    }
    return f1;
}

#if 0
static struct fml_node *fml_sub0 (Fml fml, struct fml_node *list)
{
    struct token token;
    struct fml_node *fn, *fn1;

    fml_init_token (&token, fml);
    assert (list);
    fml_cmd_lex (&list, &token);
    fn = fml_sub1 (fml, &list, &token);
    if (token.kind == '\0')
    {
        fml_del_token (&token, fml);
        return fn;
    }
    fn1 = fml_node_alloc (fml);
    fn1->p[0] = fn;
    fn = fn1;
    while (token.kind != '\0')
    {
        fn1 = fn1->p[1] = fml_node_alloc (fml);
        fn1->p[0] = fml_sub1 (fml, &list, &token);
    }
    fml_del_token (&token, fml);
    return fn;
}
#else
static struct fml_node *fml_sub0 (Fml fml, struct fml_node *list)
{
    struct token token;
    struct fml_node *fn, *fn0, *fn1;

    if (!list)
        return NULL;
    fml_init_token (&token, fml);
    fml_cmd_lex (&list, &token);
    fn1 = fn = fml_sub1 (fml, &list, &token);
    if (!fn)
    {
        fml_del_token (&token, fml);
        return fn;
    }
    if (fn->p[1] && token.kind != '\0')
    {
        fn1 = fml_node_alloc (fml);
        fn1->p[0] = fn;
    }
    fn0 = fn1;
    while (token.kind != '\0')
    {
        fn = fml_sub1 (fml, &list, &token);
        if (fn->p[1])
        {
            fn1 = fn1->p[1] = fml_node_alloc (fml);
            fn1->p[0] = fn;
        }
        else
        {
            fn1 = fn1->p[1] = fn;
        }
    }
    fml_del_token (&token, fml);
    return fn0;
}
#endif

static struct fml_node *fml_exec_foreach (struct fml_sym_info *info, Fml fml,
                                          struct fml_node **lp,
                                          struct token *tp)
{
    struct fml_sym_info *info_var;
    struct fml_node *fn, *body;
    struct fml_node *return_value = NULL, *rv;

    fml_cmd_lex (lp, tp);
    assert (tp->kind == 't');
    
    info_var = fml_sym_lookup_local (fml->sym_tab, tp->tokenbuf);
    if (!info_var)
    {
        info_var = fml_sym_add_local (fml->sym_tab, tp->tokenbuf);
        info_var->body = NULL;
        info_var->kind = FML_VAR;
    }
    else
    {
        if (info_var->kind == FML_VAR)
            fml_node_delete (fml, info_var->body);
        info_var->body = NULL;
    }
    if (fml->debug & 1)
    {
        pr_indent (1);
        printf ("[foreach %s ", tp->tokenbuf);
    }
    fml_cmd_lex (lp, tp);
    assert (tp->kind == 'g');
    fn = fml_sub0 (fml, tp->sub);
 
    fml_cmd_lex (lp, tp);
    assert (tp->kind == 'g');
    body = tp->sub;
 
    while (fn)
    {
        struct fml_node *fn1;

        fn1 = fn->p[1];
        fn->p[1] = NULL;
        if (fn->is_atom)
            info_var->body = fn;
        else
            info_var->body = fn->p[0];
        if (fml->debug & 1)
        {
            pr_indent (1);
            printf ("[foreach loop var=");
            fml_pr_list (info_var->body);
            pr_indent (-1);
        }
        rv = fml_exec_group (body, fml);
        if (rv)
            return_value = rv;
        fml_node_delete (fml, fn);
        fn = fn1;
    }
    info_var->body = NULL;
    if (fml->debug & 1)
        pr_indent (-1);
    return return_value;
}

static struct fml_node *fml_exec_if (struct fml_sym_info *info, Fml fml,
                                     struct fml_node **lp, struct token *tp)
{
    struct fml_node *fn, *body;
    struct fml_node *rv, *return_value = NULL;

    fml_cmd_lex (lp, tp);
    assert (tp->kind == 'g');
    fn = fml_sub0 (fml, tp->sub);
    fml_cmd_lex (lp, tp);
    assert (tp->kind == 'g');
    if (fn)
    {
        rv = fml_exec_group (tp->sub, fml);
        if (rv)
            return_value = rv;
    }
    fml_cmd_lex (lp, tp);
    if (tp->kind == 'e')
    {
        info = fml_sym_lookup (fml->sym_tab, tp->tokenbuf);
        if (info->kind == FML_ELSE)
        {
            fml_cmd_lex (lp, tp);
            assert (tp->kind == 'g');
            body = tp->sub;
            if (!fn)
            {
                rv = fml_exec_group (body, fml);
                if (rv)
                    return_value = rv;
            }
            fml_cmd_lex (lp, tp);
        }
    }
    fml_node_delete (fml, fn);
    return return_value;
}

static struct fml_node *fml_exec_while (struct fml_sym_info *info, Fml fml,
                                        struct fml_node **lp, struct token *tp)
{
    struct fml_node *fn, *body;
    struct fml_node *return_value = NULL;

    fml_cmd_lex (lp, tp);
    assert (tp->kind == 'g');
    fn = tp->sub;

    fml_cmd_lex (lp, tp);
    assert (tp->kind == 'g');
    body = tp->sub;
    assert (tp->sub);
    while (1)
    {
        struct fml_node *fn_expr;
        struct fml_node *rv;
        if (!fn)
            break;
        fn_expr = fml_sub0 (fml, fn);
        if (!fn_expr)
            break;
        fml_node_delete (fml, fn_expr);
        rv = fml_exec_group (body, fml);
        if (rv)
            return_value = rv;
    }
    return return_value;
}

static void fml_exec_set (struct fml_sym_info *info, Fml fml,
                    struct fml_node **lp, struct token *tp)
{
    struct fml_node *fn;
    struct fml_sym_info *info_var;
    
    fml_cmd_lex (lp, tp);
    info_var = fml_sym_lookup_local (fml->sym_tab, tp->tokenbuf);
    if (!info_var)
    {
        info_var = fml_sym_add_local (fml->sym_tab, tp->tokenbuf);
        info_var->body = NULL;
    }
    if (fml->debug & 1)
    {
        pr_indent (1);
        printf ("set %s ", tp->tokenbuf);
    }
    info_var->kind = FML_VAR;
    fml_cmd_lex (lp, tp);

    if (tp->kind == 'g')
    {
        fn = fml_sub0 (fml, tp->sub);
        fml_cmd_lex (lp, tp);
    }
    else
        fn = fml_sub2 (fml, lp, tp);
    fml_node_delete (fml, info_var->body); 
    info_var->body = fn;
    if (fml->debug & 1)
    {
        fml_pr_list (info_var->body);
        pr_indent (-1);
    }
}

static void fml_emit_expr (Fml fml, struct fml_node **lp, struct token *tp)
{
    struct fml_node *fn;

    fn = fml_sub1 (fml, lp, tp);
    fml_emit (fml, fn);
    fml_node_delete (fml, fn);
}

struct fml_node *fml_exec_group (struct fml_node *list, Fml fml)
{
    struct token token;
    struct fml_sym_info *info;
    int first = 1;
    struct fml_node *return_value = NULL, *rv;

    if (!list)
        return NULL;
    fml_init_token (&token, fml);
    fml_cmd_lex (&list, &token);
    while (token.kind)
    {
        switch (token.kind)
        {
        case 'g':
            rv = fml_exec_group (token.sub, fml);
            if (rv)
                return_value = rv;
            break;
        case 'e':
            info = fml_sym_lookup (fml->sym_tab, token.tokenbuf);
            if (info)
            {
                struct fml_node *fn;

                switch (info->kind)
                {
                case FML_FUNC:
                    fml_cmd_lex (&list, &token);
                    assert (token.kind == 't');
                    info = fml_sym_lookup (fml->sym_tab, token.tokenbuf);
                    if (!info)
                        info = fml_sym_add (fml->sym_tab, token.tokenbuf);
                    info->kind = FML_PREFIX;
                    info->args = NULL;
                    while (1)
                    {
                        fml_cmd_lex (&list, &token);
                        if (token.kind != 't' && token.kind != 'e')
                            break;
                        if (!info->args)
                        {
                            info->args = fn = fml_node_alloc (fml);
                        }
                        else
                        {
                            for (fn = info->args; fn->p[1]; fn=fn->p[1])
                                ;
                            fn = fn->p[1] = fml_node_alloc (fml);
                        }
                        fn->p[0] = token.atom;
                        fn->is_atom = 1;
                    }
                    assert (token.kind == 'g');
                    info->body = token.sub;
                    break;
                case FML_BIN:
                    fml_cmd_lex (&list, &token);
                    assert (token.kind == 't');
                    info = fml_sym_lookup (fml->sym_tab, token.tokenbuf);
                    if (!info)
                        info = fml_sym_add (fml->sym_tab, token.tokenbuf);
                    info->kind = FML_BINARY;

                    fml_cmd_lex (&list, &token);
                    assert (token.kind == 't');
                    info->args = fn = fml_node_alloc (fml);
                    fn->p[0] = token.atom;
                    fn->is_atom = 1;

                    fml_cmd_lex (&list, &token);
                    assert (token.kind == 't');
                    fn = fn->p[1] = fml_node_alloc (fml);
                    fn->p[0] = token.atom;
                    fn->is_atom = 1;

                    fml_cmd_lex (&list, &token);
                    assert (token.kind == 'g');
                    info->body = token.sub;
                    break;
                case FML_VAR:
                case FML_PREFIX:
                case FML_CPREFIX:
                    if (token.separate && !first)
                        (*fml->write_func) (' ');
                    first = 1;
                    fml_emit_expr (fml, &list, &token);
                    fml_node_stat (fml);
                    continue;
                case FML_FOREACH:
                    rv = fml_exec_foreach (info, fml, &list, &token);
                    if (rv)
                        return_value = rv;
                    break;
                case FML_IF:
                    rv = fml_exec_if (info, fml, &list, &token);
                    if (rv)
                        return_value = rv;
                    continue;
                case FML_SET:
                    fml_exec_set (info, fml, &list, &token);
                    fml_node_stat (fml);
                    continue;
                case FML_WHILE:
                    rv = fml_exec_while (info, fml, &list, &token);
                    if (rv)
                        return_value = rv;
                    break;
                case FML_RETURN:
                    fml_cmd_lex (&list, &token);

                    if (token.kind == 'g')
                    {
                        return_value = fml_sub0 (fml, token.sub);
                        fml_cmd_lex (&list, &token);
                    }
                    else
                        return_value = fml_sub2 (fml, &list, &token);
                    if (fml->debug & 1)
                    {
                        pr_indent (1);
                        printf ("return of:");
                        fml_pr_list (return_value);
                        pr_indent (-1);
                    }
                    continue;
                case FML_CODE:
                    fml_exec_group (info->body, fml);
                    break;
                default:
                    printf ("<unknown token: `%s'>", token.tokenbuf);
                    fml_cmd_lex (&list, &token);
                }
            }
            else
            {
                printf ("<unknown %s>", token.tokenbuf);
            }
            break;
        case 't':
            if (token.separate && !first)
                (*fml->write_func) (' ');
            first = 0;
            fml_emit_expr (fml, &list, &token);
            fml_node_stat (fml);
            continue;
        }
        fml_cmd_lex (&list, &token);
    }
    fml_del_token (&token, fml);
    return return_value;
}

void fml_exec (Fml fml)
{
    fml_node_stat (fml);
    fml_exec_group (fml->list, fml);
    if (fml->debug & 1)
        putchar ('\n');
}
